# forked from http.py to match icinga Rest API
from sauna.consumers.base import QueuedConsumer
from sauna.consumers import ConsumerRegister

my_consumer = ConsumerRegister('HTTP-icinga')


@my_consumer.consumer()
class HTTPIcingaConsumer(QueuedConsumer):

    def __init__(self, config):
        super().__init__(config)
        try:
            import requests
            self.requests = requests
        except ImportError:
            from ... import DependencyError
            raise DependencyError(self.__class__.__name__, 'requests',
                                  'requests', 'python3-requests')
        self.config = {
            'url': config.get('url', 'http://localhost'),
            'timeout': config.get('timeout', 60),
            'headers': config.get('headers', None)
        }

    def _send(self, service_check):
        data = {
            "filter": (
                "host.name==\"" + service_check.hostname +
                "\" && service.name==\"" + service_check.name + "\""
            ),
            "exit_status": service_check.status,
            "plugin_output": service_check.output,
            "type": "Service"
        }
        response = self.requests.post(
            self.config['url'], timeout=self.config['timeout'],
            headers=self.config['headers'], json=data
        )
        response.raise_for_status()

    @staticmethod
    def config_sample():
        return '''
        # Posts a service check trough HTTP to Icinga
        # Payload is serialized in JSON
        - type: HTTP-icinga
          url: http://icinga.host:5665/v1/actions/process-check-result
          timeout: 60
          headers:
            accept: application/json
            authorization: ICINGA_BASIC
        '''



import json
from http.server import HTTPServer, BaseHTTPRequestHandler
from logging import getLogger

from sauna.consumers.base import AsyncConsumer
from sauna.consumers import ConsumerRegister
from sauna import __version__

logger = getLogger('sauna.HTTPServerConsumer')
my_consumer = ConsumerRegister('HTTPServer')


class StoppableHTTPServer(HTTPServer):
    """HTTPServer that stops itself when receiving a threading.Event"""

    def __init__(self, must_stop, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._must_stop = must_stop

    def service_actions(self):
        """Called by the serve_forever() loop.
        Check if Sauna requested the server to shutdown.
        It cannot call self.shutdown() because the server does not run in
        a separated thread.
        """
        if self._must_stop.is_set():
            self._BaseServer__shutdown_request = True


@my_consumer.consumer()
class HTTPServerConsumer(AsyncConsumer):

    def __init__(self, config):
        super().__init__(config)
        self.config = {
            'port': config.get('port', 8080),
            'data_type': config.get('data_type', 'json'),
            'address': config.get('address', '')  # listen on all interfaces
        }

    def run(self, must_stop, *args):
        http_server = StoppableHTTPServer(
            must_stop,
            (self.config['address'], self.config['port']),
            self.HandlerFactory()
        )
        http_server.serve_forever()
        self.logger.debug('Exited consumer thread')

    @staticmethod
    def config_sample():
        return '''
        # HTTP Server that exposes sauna status
        # as a REST API or a web dashboard
        - type: HTTPServer
          port: 8080
          data_type: json # Can be json or html
        '''

    def HandlerFactory(self):
        config = self.config

        class Handler(BaseHTTPRequestHandler):

            server_version = 'Sauna/' + __version__

            def do_GET(self):
                data = self.generate_response()
                self.wfile.write(data)

            def do_HEAD(self):
                self.generate_response()

            def generate_response(self):
                try:
                    content = self.get_content_from_path()
                    code = 200
                except NotFoundError:
                    content = {'error': 'Resource not found'}
                    code = 404

                self.send_response(code)
                if config['data_type'] == 'json':
                    self.send_header('Content-Type', 'application/json')
                    data = json.dumps(content).encode()
                elif config['data_type'] == 'html':
                    self.send_header('Content-Type', 'text/html')
                    from .html import get_html
                    data = get_html()
                else:
                    data = 'data type not found'.encode()
                self.send_header('Content-Length', len(data))
                self.end_headers()
                return data

            def get_content_from_path(self):
                if self.path == '/':
                    status, code = HTTPServerConsumer.get_current_status()
                    return {
                        'status': status,
                        'code': code,
                        'checks': HTTPServerConsumer.get_checks_as_dict()
                    }
                else:
                    raise NotFoundError()

            def log_message(self, format, *args):
                logger.debug(
                    '{} {}'.format(self.address_string(), format % args))

        return Handler


class NotFoundError(Exception):
    pass



import datetime

try:
    from html.parser.HTMLParser import escape
except ImportError:
    from html import escape

from . import HTTPServerConsumer
from .dashboard_html import dashboard_html


def get_check_html():
    checks = HTTPServerConsumer.get_checks_as_dict()
    checks_html = ''
    for check_name, data in sorted(checks.items()):
        date = datetime.datetime.fromtimestamp(
            data['timestamp']
        ).strftime('%Y-%m-%d %H:%M:%S')
        checks_html += "<tr>"
        checks_html += "<td>{}</td><td><span class=\"st st_{}\">{}</span>" \
                       "</td><td>{}</td>" \
                       "<td>{}</td>".format(
                        escape(check_name), data['code'], data['status'],
                        escape(data['output']), date)
        checks_html += "</tr>"
    return checks_html


def get_html():
    template = dashboard_html
    checks_html = get_check_html()
    return template.format(checks_html).encode()



from sauna.plugins import Plugin, PluginRegister

my_plugin = PluginRegister('Apt')


@my_plugin.plugin()
class AptPlugin(Plugin):

    def __init__(self, config):
        super().__init__(config)
        try:
            import apt
            self._apt = apt
        except ImportError:
            from ... import DependencyError
            raise DependencyError(
                self.__class__.__name__,
                'apt',
                deb='python3-apt'
            )
        self._packages = None

    @property
    def packages(self) -> list:
        if self._packages is None:
            with self._apt.Cache() as cache:
                cache.upgrade()  # Only reads the packages to upgrade
                self._packages = cache.get_changes()
        return self._packages

    @my_plugin.check()
    def security_updates(self, check_config):
        num_security_packages = 0
        for p in self.packages:
            for o in p.candidate.origins:
                if 'security' in o.codename.lower():
                    num_security_packages += 1
                    break

                if 'security' in o.label.lower():
                    num_security_packages += 1
                    break

                if 'security' in o.site.lower():
                    num_security_packages += 1
                    break

        if num_security_packages == 0:
            return self.STATUS_OK, 'No security updates'

        return (
            self.STATUS_WARN,
            f'{num_security_packages} packages with security updates'
        )

    @my_plugin.check()
    def package_updates(self, check_config):
        if not self.packages:
            return self.STATUS_OK, 'No package updates'

        return (
            self.STATUS_WARN,
            f'{len(self.packages)} packages updates'
        )

    @staticmethod
    def config_sample():
        return '''
        # Debian APT
        # This only consults the local APT cache, it does not
        # run an 'apt update'. Use 'unattended-upgrades' for that.
        - type: Apt
          checks:
            - type: security_updates
            - type: package_updates
        '''



dashboard_html = """<!DOCTYPE html>
<html>
<head>
<meta charset="UTF-8">
<title>Sauna status</title>
<style>
body {{
    background-color: rgba(114, 254, 149, 0.38);
}}
h1   {{color: blue;}}
.content {{
    border-radius: 25px;
    border: 1px solid black;
    background-color: white;
    width: 80%;
    margin-left: auto;
    margin-right: auto;
}}
.st {{
    padding: 5px 10px 5px 10px;
}}
.st_0 {{
    background-color: rgba(166, 241, 166, 0.54);
    border-radius: 25px;
}}
.st_1 {{
    background-color: rgba(249, 215, 46, 0.54);
    border-radius: 25px;
}}
.st_2 {{
    background-color: rgba(249, 93, 46, 0.54);
    border-radius: 25px;
}}
.st_3 {{
    background-color: rgba(45, 149, 222, 0.54);
    border-radius: 25px;
}}
table caption {{
    font-size: 50px;
}}
table {{
    width: 80%;
    border-collapse: collapse;
    margin-left: auto;
    margin-right: auto;
}}
th, td {{
    padding: 15px;
    text-align: left;
    border-top:1pt solid black;
}}
</style>
</head>
<body>
<div class="content">
<table>
<caption>Sauna status</caption>
{}
</table>
</div>
</body>
</html>"""



import socket

from sauna.plugins import Plugin, PluginRegister

my_plugin = PluginRegister('SimpleDomain')


@my_plugin.plugin()
class SimpleDomain(Plugin):

    @my_plugin.check()
    def request(self, check_config):
        domain = check_config.get('domain')
        if check_config.get('ip_version') == 6:
            af = socket.AF_INET6
        elif check_config.get('ip_version') == 4:
            af = socket.AF_INET
        else:
            af = 0

        try:
            result = socket.getaddrinfo(domain, 0, af)
        except Exception as e:
            return Plugin.STATUS_CRIT, '{}'.format(e)

        ips = [ip[4][0] for ip in result]
        return (
            Plugin.STATUS_OK,
            'Domain was resolved with {}'.format(', '.join(ips))
        )

    @staticmethod
    def config_sample():
        return '''
        # Make a domain request,
        # crit if domain can't be resolved
        - type: SimpleDomain
          checks:
            - type: request
              domain: www.website.tld
              ip_version: 4 # 4 or 6; default is both

        '''



from functools import lru_cache
import time

from sauna.plugins.base import PsutilPlugin
from sauna.plugins import human_to_bytes, bytes_to_human, PluginRegister

my_plugin = PluginRegister('Network')


@my_plugin.plugin()
class Network(PsutilPlugin):

    def __init__(self, config):
        super().__init__(config)

    @my_plugin.check()
    def upload_data_speed(self, check_config):
        ul, _, _, _ = self.get_network_data(
            interface=check_config['interface'])
        ul = round(ul, 2)

        return (
            self._value_to_status_less(ul, check_config, human_to_bytes),
            'Upload speed: {}/s'.format(bytes_to_human(ul))
        )

    @my_plugin.check()
    def download_data_speed(self, check_config):
        _, dl, _, _ = self.get_network_data(
            interface=check_config['interface'])
        dl = round(dl, 2)
        return (
            self._value_to_status_less(dl, check_config, human_to_bytes),
            'Download speed: {}/s'.format(bytes_to_human(dl))
        )

    @my_plugin.check()
    def upload_packet_speed(self, check_config):
        _, _, ul, _ = self.get_network_data(
            interface=check_config['interface'])
        ul = round(ul, 2)
        return (
            self._value_to_status_less(ul, check_config),
            'Upload : {} p/s'.format(ul)
        )

    @my_plugin.check()
    def download_packet_speed(self, check_config):
        _, _, _, dl = self.get_network_data(
            interface=check_config['interface'])
        dl = round(dl, 2)
        return (
            self._value_to_status_less(dl, check_config),
            'Download : {} p/s'.format(dl)
        )

    @lru_cache()
    def get_network_data(self, interface='eth0', delay=1):
        t0 = time.time()
        counter = self.psutil.net_io_counters(pernic=True)[interface]
        first_values = (counter.bytes_sent, counter.bytes_recv,
                        counter.packets_sent, counter.packets_recv)

        time.sleep(delay)
        counter = self.psutil.net_io_counters(pernic=True)[interface]
        t1 = time.time()
        last_values = (counter.bytes_sent, counter.bytes_recv,
                       counter.packets_sent, counter.packets_recv)
        kb_ul, kb_dl, p_ul, p_dl = [
            (last - first) / (t1 - t0)
            for last, first in zip(last_values, first_values)
        ]
        return kb_ul, kb_dl, p_ul, p_dl

    @staticmethod
    def config_sample():
        return '''
  - type: Network

    checks:
      - type: upload_data_speed
        interface: em1
        # Crit if download > 2MB/s
        warn: 500K
        crit: 2M
      - type: download_data_speed
        interface: em1
        # Warn if upload > 500KB/s
        warn: 500K
        crit: 2M
      - type: upload_packet_speed
        interface: em1
        # Values are in packet/s
        warn: 500
        crit: 2000
      - type: download_packet_speed
        interface: em1
        # Values are in packet/s
        warn: 500
        crit: 2000
        '''



import json
import re
from typing import Optional, Set, List

from sauna.consumers.base import BatchQueuedConsumer
from sauna.consumers import ConsumerRegister

my_consumer = ConsumerRegister('HomeAssistantMQTT')


status_to_name = {
    0: 'OK',
    1: 'Warning',
    2: 'Critical',
    3: 'Unknown'
}


@my_consumer.consumer()
class HomeAssistantMQTTConsumer(BatchQueuedConsumer):
    """Report checks to Home Assistant via MQTT.

    Uses the Home Assistant MQTT discovery functionnality to
    dynamically create sensors for each couple of host/check.
    The first time Sauna sees a new check for a host, it inserts
    a MQTT message on a well known topic that HA monitors with the
    configuration for the new sensor. Each host/check gets its own
    MQTT topic that HA starts subscribing to.

    Afterwards checks are just sent on their own topic.
    """

    def __init__(self, config):
        super().__init__(config)
        try:
            import paho.mqtt.publish as mqtt_publish
            self.mqtt_publish = mqtt_publish
        except ImportError:
            from ... import DependencyError
            raise DependencyError(self.__class__.__name__, 'paho-mqtt',
                                  'paho-mqtt', 'python3-paho-mqtt')
        self.config = {
            'hostname': config.get('hostname', 'localhost'),
            'port': config.get('port', 1883),
            'auth': config.get('auth', None)
        }
        self._configured_checks: Set[str] = set()

        from ... import __version__
        self._version = __version__

    def _to_safe_str(self, text: str) -> str:
        return re.sub('[^0-9a-zA-Z]+', '_', text)

    def _get_check_discovery(self, service_check) -> Optional[dict]:
        safe_hostname = self._to_safe_str(service_check.hostname)
        safe_name = self._to_safe_str(service_check.name)
        unique_id = f"sauna_{safe_hostname}_{safe_name}"
        if unique_id in self._configured_checks:
            return None

        self._configured_checks.add(unique_id)
        state_topic = f"sauna/{safe_hostname}/{safe_name}/state"
        return {
            "topic": f"homeassistant/sensor/{unique_id}/config",
            "retain": True,
            "qos": 1,
            "payload": json.dumps({
                "value_template": "{{ value_json.status }}",
                "state_topic": state_topic,
                "json_attributes_topic": state_topic,
                "device": {
                    "identifiers": [service_check.hostname],
                    "name": service_check.hostname,
                    "manufacturer": "Sauna",
                    "sw_version": f"Sauna {self._version}"
                },
                "name": f"{service_check.hostname} {service_check.name}",
                "unique_id": unique_id,
                "platform": "mqtt"
            })
        }

    def _send_batch(self, service_checks: list):
        msgs: List[dict] = list()
        pending_discovery_checks: Set[str] = set()

        for service_check in service_checks:

            safe_hostname = self._to_safe_str(service_check.hostname)
            safe_name = self._to_safe_str(service_check.name)
            unique_id = f"sauna_{safe_hostname}_{safe_name}"

            discovery_msg = self._get_check_discovery(service_check)
            if discovery_msg is not None:
                msgs.append(discovery_msg)
                pending_discovery_checks.add(unique_id)

            msgs.append({
                'topic': f"sauna/{safe_hostname}/{safe_name}/state",
                'retain': False,
                'qos': 0,
                'payload': json.dumps({
                    "status": status_to_name.get(service_check.status),
                    "output": service_check.output
                })
            })

        try:
            self.mqtt_publish.multiple(
                msgs=msgs,
                hostname=self.config['hostname'],
                port=self.config['port'],
                auth=self.config['auth']
            )
        except Exception:
            # Remove the discovery message that could not be sent so that
            # they will be recreated the next time they are seen.
            self._configured_checks.difference_update(pending_discovery_checks)
            raise

    @staticmethod
    def config_sample():
        return '''
        # Report checks to Home Assistant via MQTT
        - type: HomeAssistantMQTT
          hostname: localhost
          port: 1883
          auth:
            username: user
            password: pass
        '''



from copy import deepcopy
from datetime import timedelta, datetime
from functools import reduce
import logging
from queue import Queue, Empty
import threading
import time


class Consumer:

    def __init__(self, config: dict):
        if config is None:
            config = {}
        self.stale_age = config.get('stale_age', 300)
        self.retry_delay = config.get('retry_delay', 10)
        self.max_retry = config.get('max_retry', -1)

    @property
    def logger(self):
        return logging.getLogger('sauna.' + self.__class__.__name__)

    @classmethod
    def logging(cls, lvl, message):
        """Log a message.

        Deprecated, use self.logger instead. Kept for backward compatibility
        """
        log = getattr(logging, lvl)
        message = '[{}] {}'.format(cls.__name__, message)
        log(message)

    def run(self, must_stop, queue):
        """Method to override in consumers."""
        raise NotImplementedError()


class BatchQueuedConsumer(Consumer):
    """Consumer that processes checks synchronously in batches.

    BatchQueuedConsumers wait for checks to appear on a queue. They buffer the
    checks until enough of them are available. They then send all checks in a
    single batch to the remote service.
    """

    #: Maximum number of checks to send in a single batch.
    max_batch_size: int = 64

    #: Maximum amount of time to wait for the batch to be full before sending
    #: it.
    max_batch_delay: timedelta = timedelta(seconds=15)

    def _send(self, service_check):
        """Send one service checks.

        Method to override in consumers sending checks one by one.
        """
        raise NotImplementedError()

    def _send_batch(self, service_checks: list):
        """Send a batch of service checks.

        Method to override in consumers for actually sending batches, otherwise
        checks are sent one by one using `self._send`.
        """
        for service_check in service_checks:
            self._send(service_check)

    def try_send(self, service_checks: list, must_stop: threading.Event):
        try:
            last_service_check = service_checks[-1]
        except IndexError:
            return

        retry_count = 0
        while True:
            retry_count = retry_count + 1

            if last_service_check.timestamp + self.stale_age < time.time():
                self.logger.warning('Dropping batch because it is too old')
                return

            if self.max_retry != -1 and retry_count > self.max_retry:
                self.logger.warning('Dropping batch because '
                                    'max_retry has been reached')
                return

            try:
                self._send_batch(service_checks)
            except Exception as e:
                self.logger.warning('Could not send batch (attempt {}/{}): {}'
                                    .format(retry_count, self.max_retry, e))
                if must_stop.is_set():
                    return

                if self.max_retry == -1 or retry_count < self.max_retry:
                    self._wait_before_retry(must_stop)
            else:
                self.logger.info('Batch sent')
                return

    def _wait_before_retry(self, must_stop: threading.Event):
        self.logger.info('Waiting %s s before retry', self.retry_delay)
        must_stop.wait(timeout=self.retry_delay)

    def run(self, must_stop, queue: Queue):
        batch = list()
        batch_created_at = datetime.utcnow()

        while not must_stop.is_set():

            # Calculate how long to wait before the current batch
            # should be sent.
            if not batch:
                wait_timeout = None
            else:
                wait_timeout = (
                    batch_created_at + self.max_batch_delay - datetime.utcnow()
                ).total_seconds()
                if wait_timeout < 0:
                    wait_timeout = 0

            try:
                service_check = queue.get(timeout=wait_timeout)
            except Empty:
                pass
            else:
                if not isinstance(service_check, threading.Event):
                    self.logger.debug('Got check {}'.format(service_check))
                    if not batch:
                        # Current batch is empty, create a new one
                        batch.append(service_check)
                        batch_created_at = datetime.utcnow()
                    else:
                        # Current batch is not empty, just append the check
                        batch.append(service_check)

            # A batch should be sent if either:
            #   - the batch isfull
            #   - the first check has waited long enough in the batch
            #   - sauna is shutting down
            should_send_batch = (
                len(batch) >= self.max_batch_size
                or
                (batch_created_at + self.max_batch_delay) < datetime.utcnow()
                or
                must_stop.is_set()
            )
            if should_send_batch:
                self.try_send(batch, must_stop)
                batch = list()

        self.logger.debug('Exited consumer thread')


class QueuedConsumer(BatchQueuedConsumer):
    """Consumer that processes checks synchronously one by one.

    QueuedConsumers wait for checks to appear on a queue. They process each
    check one by one in order until the queue is empty.
    """

    max_batch_size = 1


class AsyncConsumer(Consumer):
    """Consumer that processes checks asynchronously.

    It is up to the consumer to read the checks when it needs to. No
    queueing is made.
    """

    @classmethod
    def get_current_status(cls):
        """Get the worse status of all check results.

        :returns: (status as str, code)
        :rtype: tuple
        """
        from sauna.plugins.base import Plugin
        from sauna import check_results_lock, check_results

        def reduce_status(accumulated, update_value):
            if update_value.status > Plugin.STATUS_CRIT:
                return accumulated
            return accumulated if accumulated > update_value.status else \
                update_value.status

        with check_results_lock:
            code = reduce(reduce_status, check_results.values(), 0)

        return Plugin.status_code_to_str(code), code

    @classmethod
    def get_checks_as_dict(cls):
        from sauna.plugins.base import Plugin
        from sauna import check_results_lock, check_results

        checks = {}
        with check_results_lock:
            for service_check in check_results.values():
                checks[service_check.name] = {
                    'status': Plugin.status_code_to_str(service_check.status),
                    'code': service_check.status,
                    'timestamp': service_check.timestamp,
                    'output': service_check.output
                }
        return deepcopy(checks)
